@@ -39,6 +39,7 @@ import (
 	"github.com/prometheus/prometheus/tsdb/fileutil"
 	"github.com/prometheus/prometheus/tsdb/tombstones"
 	"github.com/prometheus/prometheus/tsdb/tsdbutil"
+	"github.com/prometheus/prometheus/tsdb/wlog"
 )
 
 func TestSplitByRange(t *testing.T) {
@@ -441,7 +442,7 @@ func TestCompactionFailWillCleanUpTempDir(t *testing.T) {
 
 	tmpdir := t.TempDir()
 
-	require.Error(t, compactor.write(tmpdir, &BlockMeta{}, erringBReader{}))
+	require.Error(t, compactor.write(tmpdir, &BlockMeta{}, DefaultBlockPopulator{}, erringBReader{}))
 	_, err = os.Stat(filepath.Join(tmpdir, BlockMeta{}.ULID.String()) + tmpForCreationBlockDirSuffix)
 	require.True(t, os.IsNotExist(err), "directory is not cleaned up")
 }
@@ -467,8 +468,8 @@ func (erringBReader) Size() int64                            { return 0 }
 
 type nopChunkWriter struct{}
 
-func (nopChunkWriter) WriteChunks(chunks ...chunks.Meta) error { return nil }
-func (nopChunkWriter) Close() error                            { return nil }
+func (nopChunkWriter) WriteChunks(...chunks.Meta) error { return nil }
+func (nopChunkWriter) Close() error                     { return nil }
 
 func samplesForRange(minTime, maxTime int64, maxSamplesPerChunk int) (ret [][]sample) {
 	var curr []sample
@@ -953,7 +954,8 @@ func TestCompaction_populateBlock(t *testing.T) {
 			}
 
 			iw := &mockIndexWriter{}
-			err = c.populateBlock(blocks, meta, iw, nopChunkWriter{})
+			blockPopulator := DefaultBlockPopulator{}
+			err = blockPopulator.PopulateBlock(c.ctx, c.metrics, c.logger, c.chunkPool, c.mergeFunc, blocks, meta, iw, nopChunkWriter{})
 			if tc.expErr != nil {
 				require.Error(t, err)
 				require.Equal(t, tc.expErr.Error(), err.Error())
@@ -974,7 +976,7 @@ func TestCompaction_populateBlock(t *testing.T) {
 						s       sample
 					)
 					for iter.Next() == chunkenc.ValFloat {
-						s.t, s.v = iter.At()
+						s.t, s.f = iter.At()
 						if firstTs == math.MaxInt64 {
 							firstTs = s.t
 						}
@@ -1171,12 +1173,9 @@ func TestCancelCompactions(t *testing.T) {
 	createBlock(t, tmpdir, genSeries(1, 1, 2000, 2001)) // The most recent block is ignored so can be e small one.
 
 	// Copy the db so we have an exact copy to compare compaction times.
-	tmpdirCopy := tmpdir + "Copy"
+	tmpdirCopy := t.TempDir()
 	err := fileutil.CopyDirs(tmpdir, tmpdirCopy)
 	require.NoError(t, err)
-	defer func() {
-		require.NoError(t, os.RemoveAll(tmpdirCopy))
-	}()
 
 	// Measure the compaction time without interrupting it.
 	var timeCompactionUninterrupted time.Duration
@@ -1184,15 +1183,14 @@ func TestCancelCompactions(t *testing.T) {
 		db, err := open(tmpdir, log.NewNopLogger(), nil, DefaultOptions(), []int64{1, 2000}, nil)
 		require.NoError(t, err)
 		require.Equal(t, 3, len(db.Blocks()), "initial block count mismatch")
-		require.Equal(t, 0.0, prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.ran), "initial compaction counter mismatch")
+		require.Equal(t, 0.0, prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.Ran), "initial compaction counter mismatch")
 		db.compactc <- struct{}{} // Trigger a compaction.
-		var start time.Time
-		for prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.populatingBlocks) <= 0 {
+		for prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.PopulatingBlocks) <= 0 {
 			time.Sleep(3 * time.Millisecond)
 		}
-		start = time.Now()
 
-		for prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.ran) != 1 {
+		start := time.Now()
+		for prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.Ran) != 1 {
 			time.Sleep(3 * time.Millisecond)
 		}
 		timeCompactionUninterrupted = time.Since(start)
@@ -1204,23 +1202,32 @@ func TestCancelCompactions(t *testing.T) {
 		db, err := open(tmpdirCopy, log.NewNopLogger(), nil, DefaultOptions(), []int64{1, 2000}, nil)
 		require.NoError(t, err)
 		require.Equal(t, 3, len(db.Blocks()), "initial block count mismatch")
-		require.Equal(t, 0.0, prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.ran), "initial compaction counter mismatch")
+		require.Equal(t, 0.0, prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.Ran), "initial compaction counter mismatch")
 		db.compactc <- struct{}{} // Trigger a compaction.
-		dbClosed := make(chan struct{})
 
-		for prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.populatingBlocks) <= 0 {
+		for prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.PopulatingBlocks) <= 0 {
 			time.Sleep(3 * time.Millisecond)
 		}
-		go func() {
-			require.NoError(t, db.Close())
-			close(dbClosed)
-		}()
 
 		start := time.Now()
-		<-dbClosed
+		require.NoError(t, db.Close())
 		actT := time.Since(start)
-		expT := time.Duration(timeCompactionUninterrupted / 2) // Closing the db in the middle of compaction should less than half the time.
+
+		expT := timeCompactionUninterrupted / 2 // Closing the db in the middle of compaction should less than half the time.
 		require.True(t, actT < expT, "closing the db took more than expected. exp: <%v, act: %v", expT, actT)
+
+		// Make sure that no blocks were marked as compaction failed.
+		// This checks that the `context.Canceled` error is properly checked at all levels:
+		// - tsdb_errors.NewMulti() should have the Is() method implemented for correct checks.
+		// - callers should check with errors.Is() instead of ==.
+		readOnlyDB, err := OpenDBReadOnly(tmpdirCopy, log.NewNopLogger())
+		require.NoError(t, err)
+		blocks, err := readOnlyDB.Blocks()
+		require.NoError(t, err)
+		for i, b := range blocks {
+			require.Falsef(t, b.Meta().Compaction.Failed, "block %d (%s) should not be marked as compaction failed", i, b.Meta().ULID)
+		}
+		require.NoError(t, readOnlyDB.Close())
 	}
 }
 
@@ -1279,7 +1286,7 @@ func TestDeleteCompactionBlockAfterFailedReload(t *testing.T) {
 			require.NoError(t, os.RemoveAll(lastBlockIndex)) // Corrupt the block by removing the index file.
 
 			require.Equal(t, 0.0, prom_testutil.ToFloat64(db.metrics.reloadsFailed), "initial 'failed db reloadBlocks' count metrics mismatch")
-			require.Equal(t, 0.0, prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.ran), "initial `compactions` count metric mismatch")
+			require.Equal(t, 0.0, prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.Ran), "initial `compactions` count metric mismatch")
 			require.Equal(t, 0.0, prom_testutil.ToFloat64(db.metrics.compactionsFailed), "initial `compactions failed` count metric mismatch")
 
 			// Do the compaction and check the metrics.
@@ -1287,7 +1294,7 @@ func TestDeleteCompactionBlockAfterFailedReload(t *testing.T) {
 			// the new block created from the compaction should be deleted.
 			require.Error(t, db.Compact())
 			require.Equal(t, 1.0, prom_testutil.ToFloat64(db.metrics.reloadsFailed), "'failed db reloadBlocks' count metrics mismatch")
-			require.Equal(t, 1.0, prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.ran), "`compaction` count metric mismatch")
+			require.Equal(t, 1.0, prom_testutil.ToFloat64(db.compactor.(*LeveledCompactor).metrics.Ran), "`compaction` count metric mismatch")
 			require.Equal(t, 1.0, prom_testutil.ToFloat64(db.metrics.compactionsFailed), "`compactions failed` count metric mismatch")
 
 			actBlocks, err = blockDirs(db.Dir())
@@ -1298,105 +1305,128 @@ func TestDeleteCompactionBlockAfterFailedReload(t *testing.T) {
 }
 
 func TestHeadCompactionWithHistograms(t *testing.T) {
-	head, _ := newTestHead(t, DefaultBlockDuration, false, false)
-	require.NoError(t, head.Init(0))
-	t.Cleanup(func() {
-		require.NoError(t, head.Close())
-	})
+	for _, floatTest := range []bool{true, false} {
+		t.Run(fmt.Sprintf("float=%t", floatTest), func(t *testing.T) {
+			head, _ := newTestHead(t, DefaultBlockDuration, wlog.CompressionNone, false)
+			require.NoError(t, head.Init(0))
+			t.Cleanup(func() {
+				require.NoError(t, head.Close())
+			})
 
-	minute := func(m int) int64 { return int64(m) * time.Minute.Milliseconds() }
-	ctx := context.Background()
-	appendHistogram := func(lbls labels.Labels, from, to int, h *histogram.Histogram, exp *[]tsdbutil.Sample) {
-		t.Helper()
-		app := head.Appender(ctx)
-		for tsMinute := from; tsMinute <= to; tsMinute++ {
-			_, err := app.AppendHistogram(0, lbls, minute(tsMinute), h)
+			minute := func(m int) int64 { return int64(m) * time.Minute.Milliseconds() }
+			ctx := context.Background()
+			appendHistogram := func(
+				lbls labels.Labels, from, to int, h *histogram.Histogram, exp *[]tsdbutil.Sample,
+			) {
+				t.Helper()
+				app := head.Appender(ctx)
+				for tsMinute := from; tsMinute <= to; tsMinute++ {
+					var err error
+					if floatTest {
+						_, err = app.AppendHistogram(0, lbls, minute(tsMinute), nil, h.ToFloat())
+						efh := h.ToFloat()
+						if tsMinute == from {
+							efh.CounterResetHint = histogram.UnknownCounterReset
+						} else {
+							efh.CounterResetHint = histogram.NotCounterReset
+						}
+						*exp = append(*exp, sample{t: minute(tsMinute), fh: efh})
+					} else {
+						_, err = app.AppendHistogram(0, lbls, minute(tsMinute), h, nil)
+						eh := h.Copy()
+						if tsMinute == from {
+							eh.CounterResetHint = histogram.UnknownCounterReset
+						} else {
+							eh.CounterResetHint = histogram.NotCounterReset
+						}
+						*exp = append(*exp, sample{t: minute(tsMinute), h: eh})
+					}
+					require.NoError(t, err)
+				}
+				require.NoError(t, app.Commit())
+			}
+			appendFloat := func(lbls labels.Labels, from, to int, exp *[]tsdbutil.Sample) {
+				t.Helper()
+				app := head.Appender(ctx)
+				for tsMinute := from; tsMinute <= to; tsMinute++ {
+					_, err := app.Append(0, lbls, minute(tsMinute), float64(tsMinute))
+					require.NoError(t, err)
+					*exp = append(*exp, sample{t: minute(tsMinute), f: float64(tsMinute)})
+				}
+				require.NoError(t, app.Commit())
+			}
+
+			var (
+				series1                = labels.FromStrings("foo", "bar1")
+				series2                = labels.FromStrings("foo", "bar2")
+				series3                = labels.FromStrings("foo", "bar3")
+				series4                = labels.FromStrings("foo", "bar4")
+				exp1, exp2, exp3, exp4 []tsdbutil.Sample
+			)
+			h := &histogram.Histogram{
+				Count:         11,
+				ZeroCount:     4,
+				ZeroThreshold: 0.001,
+				Sum:           35.5,
+				Schema:        1,
+				PositiveSpans: []histogram.Span{
+					{Offset: 0, Length: 2},
+					{Offset: 2, Length: 2},
+				},
+				PositiveBuckets: []int64{1, 1, -1, 0},
+				NegativeSpans: []histogram.Span{
+					{Offset: 0, Length: 1},
+					{Offset: 1, Length: 2},
+				},
+				NegativeBuckets: []int64{1, 2, -1},
+			}
+
+			// Series with only histograms.
+			appendHistogram(series1, 100, 105, h, &exp1)
+
+			// Series starting with float and then getting histograms.
+			appendFloat(series2, 100, 102, &exp2)
+			appendHistogram(series2, 103, 105, h.Copy(), &exp2)
+			appendFloat(series2, 106, 107, &exp2)
+			appendHistogram(series2, 108, 109, h.Copy(), &exp2)
+
+			// Series starting with histogram and then getting float.
+			appendHistogram(series3, 101, 103, h.Copy(), &exp3)
+			appendFloat(series3, 104, 106, &exp3)
+			appendHistogram(series3, 107, 108, h.Copy(), &exp3)
+			appendFloat(series3, 109, 110, &exp3)
+
+			// A float only series.
+			appendFloat(series4, 100, 102, &exp4)
+
+			// Compaction.
+			mint := head.MinTime()
+			maxt := head.MaxTime() + 1 // Block intervals are half-open: [b.MinTime, b.MaxTime).
+			compactor, err := NewLeveledCompactor(context.Background(), nil, nil, []int64{DefaultBlockDuration}, chunkenc.NewPool(), nil)
 			require.NoError(t, err)
-			*exp = append(*exp, sample{t: minute(tsMinute), h: h.Copy()})
-		}
-
-		require.NoError(t, app.Commit())
-	}
-	appendFloat := func(lbls labels.Labels, from, to int, exp *[]tsdbutil.Sample) {
-		t.Helper()
-		app := head.Appender(ctx)
-		for tsMinute := from; tsMinute <= to; tsMinute++ {
-			_, err := app.Append(0, lbls, minute(tsMinute), float64(tsMinute))
+			id, err := compactor.Write(head.opts.ChunkDirRoot, head, mint, maxt, nil)
 			require.NoError(t, err)
-			*exp = append(*exp, sample{t: minute(tsMinute), v: float64(tsMinute)})
-		}
-		require.NoError(t, app.Commit())
+			require.NotEqual(t, ulid.ULID{}, id)
+
+			// Open the block and query it and check the histograms.
+			block, err := OpenBlock(nil, path.Join(head.opts.ChunkDirRoot, id.String()), nil)
+			require.NoError(t, err)
+			t.Cleanup(func() {
+				require.NoError(t, block.Close())
+			})
+
+			q, err := NewBlockQuerier(block, block.MinTime(), block.MaxTime())
+			require.NoError(t, err)
+
+			actHists := query(t, q, labels.MustNewMatcher(labels.MatchRegexp, "foo", "bar.*"))
+			require.Equal(t, map[string][]tsdbutil.Sample{
+				series1.String(): exp1,
+				series2.String(): exp2,
+				series3.String(): exp3,
+				series4.String(): exp4,
+			}, actHists)
+		})
 	}
-
-	var (
-		series1                = labels.FromStrings("foo", "bar1")
-		series2                = labels.FromStrings("foo", "bar2")
-		series3                = labels.FromStrings("foo", "bar3")
-		series4                = labels.FromStrings("foo", "bar4")
-		exp1, exp2, exp3, exp4 []tsdbutil.Sample
-	)
-	h := &histogram.Histogram{
-		Count:         11,
-		ZeroCount:     4,
-		ZeroThreshold: 0.001,
-		Sum:           35.5,
-		Schema:        1,
-		PositiveSpans: []histogram.Span{
-			{Offset: 0, Length: 2},
-			{Offset: 2, Length: 2},
-		},
-		PositiveBuckets: []int64{1, 1, -1, 0},
-		NegativeSpans: []histogram.Span{
-			{Offset: 0, Length: 1},
-			{Offset: 1, Length: 2},
-		},
-		NegativeBuckets: []int64{1, 2, -1},
-	}
-
-	// Series with only histograms.
-	appendHistogram(series1, 100, 105, h, &exp1)
-
-	// Series starting with float and then getting histograms.
-	appendFloat(series2, 100, 102, &exp2)
-	appendHistogram(series2, 103, 105, h.Copy(), &exp2)
-	appendFloat(series2, 106, 107, &exp2)
-	appendHistogram(series2, 108, 109, h.Copy(), &exp2)
-
-	// Series starting with histogram and then getting float.
-	appendHistogram(series3, 101, 103, h.Copy(), &exp3)
-	appendFloat(series3, 104, 106, &exp3)
-	appendHistogram(series3, 107, 108, h.Copy(), &exp3)
-	appendFloat(series3, 109, 110, &exp3)
-
-	// A float only series.
-	appendFloat(series4, 100, 102, &exp4)
-
-	// Compaction.
-	mint := head.MinTime()
-	maxt := head.MaxTime() + 1 // Block intervals are half-open: [b.MinTime, b.MaxTime).
-	compactor, err := NewLeveledCompactor(context.Background(), nil, nil, []int64{DefaultBlockDuration}, chunkenc.NewPool(), nil)
-	require.NoError(t, err)
-	id, err := compactor.Write(head.opts.ChunkDirRoot, head, mint, maxt, nil)
-	require.NoError(t, err)
-	require.NotEqual(t, ulid.ULID{}, id)
-
-	// Open the block and query it and check the histograms.
-	block, err := OpenBlock(nil, path.Join(head.opts.ChunkDirRoot, id.String()), nil)
-	require.NoError(t, err)
-	t.Cleanup(func() {
-		require.NoError(t, block.Close())
-	})
-
-	q, err := NewBlockQuerier(block, block.MinTime(), block.MaxTime())
-	require.NoError(t, err)
-
-	actHists := query(t, q, labels.MustNewMatcher(labels.MatchRegexp, "foo", "bar.*"))
-	require.Equal(t, map[string][]tsdbutil.Sample{
-		series1.String(): exp1,
-		series2.String(): exp2,
-		series3.String(): exp3,
-		series4.String(): exp4,
-	}, actHists)
 }
 
 // Depending on numSeriesPerSchema, it can take few gigs of memory;
@@ -1423,12 +1453,6 @@ func TestSparseHistogramSpaceSavings(t *testing.T) {
 		{100, 15, 3, 5},
 		{100, 50, 3, 3},
 		{100, 100, 3, 2},
-		//{1000, 15, 1, 0},
-		//{1000, 50, 1, 0},
-		//{1000, 100, 1, 0},
-		//{1000, 15, 3, 5},
-		//{1000, 50, 3, 3},
-		//{1000, 100, 3, 2},
 	}
 
 	type testSummary struct {
@@ -1462,11 +1486,11 @@ func TestSparseHistogramSpaceSavings(t *testing.T) {
 				c.numBuckets,
 			),
 			func(t *testing.T) {
-				oldHead, _ := newTestHead(t, DefaultBlockDuration, false, false)
+				oldHead, _ := newTestHead(t, DefaultBlockDuration, wlog.CompressionNone, false)
 				t.Cleanup(func() {
 					require.NoError(t, oldHead.Close())
 				})
-				sparseHead, _ := newTestHead(t, DefaultBlockDuration, false, false)
+				sparseHead, _ := newTestHead(t, DefaultBlockDuration, wlog.CompressionNone, false)
 				t.Cleanup(func() {
 					require.NoError(t, sparseHead.Close())
 				})
@@ -1478,11 +1502,11 @@ func TestSparseHistogramSpaceSavings(t *testing.T) {
 
 				for sid, schema := range allSchemas {
 					for i := 0; i < c.numSeriesPerSchema; i++ {
-						lbls := labels.Labels{
-							{Name: "__name__", Value: fmt.Sprintf("rpc_durations_%d_histogram_seconds", i)},
-							{Name: "instance", Value: "localhost:8080"},
-							{Name: "job", Value: fmt.Sprintf("sparse_histogram_schema_%s", schemaDescription[sid])},
-						}
+						lbls := labels.FromStrings(
+							"__name__", fmt.Sprintf("rpc_durations_%d_histogram_seconds", i),
+							"instance", "localhost:8080",
+							"job", fmt.Sprintf("sparse_histogram_schema_%s", schemaDescription[sid]),
+						)
 						allSparseSeries = append(allSparseSeries, struct {
 							baseLabels labels.Labels
 							hists      []*histogram.Histogram
@@ -1511,7 +1535,7 @@ func TestSparseHistogramSpaceSavings(t *testing.T) {
 						)
 						for i := 0; i < numHistograms; i++ {
 							ts := int64(i) * timeStep
-							ref, err = sparseApp.AppendHistogram(ref, ah.baseLabels, ts, ah.hists[i])
+							ref, err = sparseApp.AppendHistogram(ref, ah.baseLabels, ts, ah.hists[i], nil)
 							require.NoError(t, err)
 						}
 					}
@@ -1546,21 +1570,20 @@ func TestSparseHistogramSpaceSavings(t *testing.T) {
 							for it.Next() {
 								numOldSeriesPerHistogram++
 								b := it.At()
-								lbls := append(ah.baseLabels, labels.Label{Name: "le", Value: fmt.Sprintf("%.16f", b.Upper)})
+								lbls := labels.NewBuilder(ah.baseLabels).Set("le", fmt.Sprintf("%.16f", b.Upper)).Labels()
 								refs[itIdx], err = oldApp.Append(refs[itIdx], lbls, ts, float64(b.Count))
 								require.NoError(t, err)
 								itIdx++
 							}
+							baseName := ah.baseLabels.Get(labels.MetricName)
 							// _count metric.
-							countLbls := ah.baseLabels.Copy()
-							countLbls[0].Value = countLbls[0].Value + "_count"
+							countLbls := labels.NewBuilder(ah.baseLabels).Set(labels.MetricName, baseName+"_count").Labels()
 							_, err = oldApp.Append(0, countLbls, ts, float64(h.Count))
 							require.NoError(t, err)
 							numOldSeriesPerHistogram++
 
 							// _sum metric.
-							sumLbls := ah.baseLabels.Copy()
-							sumLbls[0].Value = sumLbls[0].Value + "_sum"
+							sumLbls := labels.NewBuilder(ah.baseLabels).Set(labels.MetricName, baseName+"_sum").Labels()
 							_, err = oldApp.Append(0, sumLbls, ts, h.Sum)
 							require.NoError(t, err)
 							numOldSeriesPerHistogram++
